package com.perforce.hive.connection;

import com.perforce.hive.ErrorModel;
import com.perforce.hive.navigation.NavBar;
import com.perforce.p4java.admin.IProtectionEntry;
import com.perforce.p4java.exception.P4JavaException;
import com.perforce.p4java.option.server.GetProtectionEntriesOptions;
import com.perforce.p4java.server.IOptionsServer;
import com.perforce.p4java.server.callback.ICommandCallback;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import spark.ModelAndView;
import spark.Request;
import spark.Response;

import java.util.Base64;
import java.util.List;
import java.util.Map;

public class ConnectionSession extends ErrorModel {

	private static Logger logger = LoggerFactory.getLogger(ConnectionSession.class);

	public static String TICKET = "ticket";
	public static String USER = "username";
	static String PASS = "password";

	public static boolean basicAuth(Request request, Response response) {
		String authHeader = request.raw().getHeader("Authorization");

		// exit early if not Basic Auth
		if (authHeader == null || !authHeader.startsWith("Basic ")) {
			return false;
		}

		String baseHeader = authHeader.substring("basic ".length());
		byte[] bytes = Base64.getDecoder().decode(baseHeader);
		String decode = new String(bytes);
		String bits[] = decode.split(":");

		if (bits.length != 2) {
			logger.info("[login] Basic Auth: no user or pass.");
			return false;
		}

		String user = bits[0];
		String pass = bits[1];

		logger.info("[login] Basic Auth: " + user);

		String ticket;
		try {
			ticket = login(user, pass);
		} catch (P4JavaException e) {
			return false;
		}
		request.session().attribute(ConnectionSession.USER, user);
		request.session().attribute(ConnectionSession.TICKET, ticket);
		return true;
	}

	public static IOptionsServer get(Request request, Response response) throws P4JavaException {
		// use Perforce ticket from session
		String user = request.session().attribute(ConnectionSession.USER);
		String ticket = request.session().attribute(ConnectionSession.TICKET);

		try {
			if (user != null) {
				logger.info("[login] Session: " + user);
				IOptionsServer p4 = getP4(user, ticket);
				return p4;
			}
		} catch (P4JavaException e) {
			// wipe out cookie and session
			logout(request, response);
		}

		// use the 'anonymous' account to view or proxy files
		user = ConnectionConfig.get().getAnonymousUser();
		ticket = ConnectionConfig.get().getAnonymousTicket();

		try {
			logger.info("[login] Anonymous");
			IOptionsServer p4 = getP4(user, ticket);
			return p4;
		} catch (P4JavaException e) {
			// wipe out cookie and session
			logout(request, response);
			throw e;
		}
	}

	public static AuthLevel getAuth(Request request) {
		String user = request.session().attribute(ConnectionSession.USER);
		String ticket = request.session().attribute(ConnectionSession.TICKET);

		// if no user fall back to Anonymous and return AuthLevel.ANONYMOUS
		if (user == null) {
			return AuthLevel.ANONYMOUS;
		}

		try {
			IOptionsServer p4 = getP4(user, ticket);
			GetProtectionEntriesOptions opts = new GetProtectionEntriesOptions();

			List<IProtectionEntry> protect = p4.getProtectionEntries(null, opts);
			for (IProtectionEntry p : protect) {
				if ("super".equals(p.getMode())) {
					return AuthLevel.SUPER;
				}
			}
			return AuthLevel.VALID;
		} catch (P4JavaException e) {
			return AuthLevel.NONE;
		}
	}

	public static IOptionsServer getP4(String user, String ticket) throws P4JavaException {
		String uri = ConnectionConfig.get().toUri();
		IOptionsServer p4 = ConnectionFactory.getConnection(uri);
		p4.connect();
		p4.setUserName(user);
		p4.setAuthTicket(ticket);

		// Register logging callback
		ICommandCallback logging = new ConnectionLogging(p4);
		p4.registerCallback(logging);

		// verify connection
		try {
			isLogin(p4);
		} catch (P4JavaException e) {
			throw e;
		}
		return p4;
	}

	public static String login(String user, String pass) throws P4JavaException {
		String uri = ConnectionConfig.get().toUri();
		IOptionsServer p4 = ConnectionFactory.getConnection(uri);

		// fail if server is down.
		p4.connect();

		// login and use Perforce ticket for session
		p4.setUserName(user);

		// fetch ticket
		String ticket = null;
		p4.login(pass);

		// check for null ticket
		ticket = p4.getAuthTicket();
		if (ticket == null) {
			throw new P4JavaException("Ticket value is null.");
		}

		return ticket;
	}

	public static ModelAndView login(Request request, Response response) {
		Map<String, Object> model;
		model = NavBar.attributes("cover", request);

		String user = request.queryParams(USER);
		String pass = request.queryParams(PASS);

		String ticket = null;
		try {
			ticket = login(user, pass);
		} catch (Exception e) {
			return error(e);
		}

		// store ticket in session
		request.session().attribute(USER, user);
		request.session().attribute(TICKET, ticket);

		// save cookie
		String store = request.queryParams("store");
		if (store != null) {
			response.cookie(USER, user);
			response.cookie(TICKET, ticket);
		}
		response.redirect("/");

		return new ModelAndView(model, "page-cover.html");
	}

	public static void logout(Request request, Response response) {
		request.session().removeAttribute(USER);
		request.session().removeAttribute(TICKET);

		response.removeCookie(USER);
		response.removeCookie(TICKET);
	}

	public static void cookie(Request request, Response response) {
		String user = request.cookie(USER);
		String ticket = request.cookie(TICKET);

		// exit early if no cookie
		if (user == null) {
			return;
		}

		logger.info("[login] Cookie: " + user);

		try {
			getP4(user, ticket);
		} catch (P4JavaException e) {
			logout(request, response);
			return;
		}

		// store ticket in session
		request.session().attribute(USER, user);
		request.session().attribute(TICKET, ticket);
	}

	private static void isLogin(IOptionsServer p4) throws P4JavaException {
		String status = p4.getLoginStatus();
		if (status.contains("not necessary")) {
			return;
		}
		if (status.contains("ticket expires in")) {
			return;
		}
		// If there is a broker or something else that swallows the message
		if (status.isEmpty()) {
			return;
		}
		throw new P4JavaException(status);
	}
}
