package com.perforce.hive.plugin.docker;

import java.io.IOException;
import java.util.List;
import java.util.Map;
import java.util.UUID;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import com.perforce.hive.ErrorModel;
import com.perforce.hive.PluginArtifactFactory;
import com.perforce.hive.api.DeployInterface;
import com.perforce.hive.api.artifact.ArtifactInterface;
import com.perforce.hive.api.artifact.ArtifactType;
import com.perforce.hive.common.JsonUtil;
import com.perforce.hive.connection.ConnectionSession;
import com.perforce.hive.depot.DepotDomain;
import com.perforce.hive.depot.DepotFactory;
import com.perforce.hive.navigation.NavBar;
import com.perforce.hive.plugin.docker.model.DigestUtil;
import com.perforce.hive.plugin.docker.model.DockerBlob;
import com.perforce.hive.plugin.docker.model.DockerBlobFactory;
import com.perforce.hive.plugin.docker.model.DockerManifest1;
import com.perforce.p4java.exception.P4JavaException;
import com.perforce.p4java.server.IOptionsServer;

import freemarker.cache.ClassTemplateLoader;
import freemarker.cache.TemplateLoader;
import freemarker.template.Configuration;
import ro.fortsoft.pf4j.Extension;
import ro.fortsoft.pf4j.PluginWrapper;
import spark.ModelAndView;
import spark.Request;
import spark.Response;
import spark.Spark;
import spark.template.freemarker.FreeMarkerEngine;

public class DockerEndPoint extends DockerPlugin {

	private static Logger logger = LoggerFactory.getLogger(DockerEndPoint.class);

	public DockerEndPoint(PluginWrapper wrapper) {
		super(wrapper);
	}

	@Extension
	public static class MavenDeploy implements DeployInterface {

		@Override
		public TemplateLoader getTemplateLoader() throws IOException {
			return new ClassTemplateLoader(DockerEndPoint.class, "");
		}

		@Override
		public void addDeployEndPoints(String root, Configuration configuration) {

			String path = root + "/" + getType().getName();

			Spark.get(path, (request, response) -> {
				return DockerEndPoint.getDeploy(request, response);
			}, new FreeMarkerEngine(configuration));

			Spark.post(path, (request, response) -> {
				return DockerEndPoint.postDeploy(request, response);
			});
		}

		@Override
		public void addAPIEndPoints(String root, Configuration configuration) {

			// Register domains for each Docker depots
			DepotDomain reg = DepotDomain.getInstance();
			reg.registerDomain(getType(), "/api/docker/");

			// Register endpoints
			String base = root + "/" + getType().getName() + "/";
			// String path = base + ":repo";
			String path = "";

			Spark.get(path + "/v2/", (request, response) -> {

				response.header("Docker-Distribution-API-Version", "registry/2.0");
				response.status(200);
				return "";
			});

			Spark.get(path + "/v2/:name/manifests/:reference", (request, response) -> {

				String repo = "docker"; // request.params(":repo");
				String name = request.params(":name");
				String ref = request.params(":reference");
				String contentType = request.headers("Accept");
				int version = getManifestVersion(contentType);

				DockerBlob file = DockerBlobFactory.get(repo, name, "manifest.v" + version);
				String json = file.read();

				response.header("Content-Length", String.valueOf(json.length()));
				return json;
			});

			Spark.head(path + "/v2/:name/manifests/:reference", (request, response) -> {

				String name = request.params(":name");
				String ref = request.params(":reference");

				DockerManifest1 manifest = new DockerManifest1();
				String json = JsonUtil.toJson(manifest);
				int length = json.length();
				String digest = DigestUtil.sha256(json);

				response.header("Content-Length", String.valueOf(length));
				response.header("Docker-Content-Digest", digest);
				response.status(200);
				return "";
			});

			// Pushing an Image Manifest
			Spark.put(path + "/v2/:name/manifests/:reference", (request, response) -> {

				String repo = "docker"; // request.params(":repo");
				String name = request.params(":name");
				String ref = request.params(":reference");
				String contentType = request.headers("Content-Type");
				int version = getManifestVersion(contentType);

				DockerBlob file = DockerBlobFactory.get(repo, name, "manifest.v" + version);
				file.write(request.bodyAsBytes());

				// Submit changes
				// get type from request parameters
				ArtifactType type = getType();

				ArtifactInterface artifact = PluginArtifactFactory.getArtifact(type, request, response);
				List<String> upload = artifact.populateMultiPartPost();

				int pending = artifact.updatePending(upload);
				artifact.submitPending(pending, "Hive Direct Deploy");
				artifact.index();
				artifact.disconnect();

				response.status(200);
				return "";
			});

			Spark.get(path + "/v2/:name/blobs/:tarsum", (request, response) -> {
				return "";
			});

			// Starting An Upload POST --> GET
			Spark.post(path + "/v2/:name/blobs/uploads/", (request, response) -> {

				String repo = "docker"; // request.params(":repo");
				String name = request.params(":name");

				// TODO upload blob and provide uuid
				UUID id = UUID.randomUUID();
				String uuid = id.toString();
				int offset = 0;
				int length = 0;

				// Clean up file in workspace
				DockerBlobFactory.clean(repo, name, uuid);

				String location = "/v2/" + name + "/blobs/uploads/" + uuid;
				response.header("Location", location);
				response.header("Range", "0-" + offset);
				response.header("Content-Length", String.valueOf(length));
				response.header("Docker-Upload-UUID", uuid);
				response.status(202);
				return "";
			});

			// Existing Layers
			Spark.head(path + "/v2/:name/blobs/:digest", (request, response) -> {

				String repo = "docker"; // request.params(":repo");
				String name = request.params(":name");
				String digest = request.params(":digest");

				DockerBlob blob = DockerBlobFactory.get(repo, name, digest);

				if (blob.exists()) {
					long length = blob.getLength();
					response.header("Content-Length", String.valueOf(length));
					response.header("Docker-Content-Digest", digest);
					response.status(200);
					return "";
				}

				else {
					response.header("Docker-Content-Digest", digest);
					response.status(404);
					return "";
				}
			});

			// Upload Progress
			Spark.get(path + "/v2/:name/blobs/uploads/:uuid", (request, response) -> {
				String host = request.headers("Host");

				String repo = request.params(":repo");
				String name = request.params(":name");
				String uuid = request.params(":uuid");

				// TODO get upload progress
				int offset = 0;

				String location = "/v2/" + name + "/blobs/uploads/" + uuid;
				response.header("Location", location);
				response.header("Range", "bytes=0-" + offset);
				response.header("Docker-Upload-UUID", uuid);
				response.status(204);
				return "";
			});

			// Chunked Upload
			Spark.patch(path + "/v2/:name/blobs/uploads/:uuid", (request, response) -> {

				String repo = "docker"; // request.params(":repo");
				String name = request.params(":name");
				String uuid = request.params(":uuid");

				DockerBlob blob = DockerBlobFactory.get(repo, name, uuid);
				long start = blob.getOffset();
				blob.write(request.bodyAsBytes());
				long end = blob.getOffset();

				String location = "/v2/" + name + "/blobs/uploads/" + uuid;
				response.header("Location", location);
				response.header("Range", start + "-" + end);
				response.header("Content-Length", "0");
				response.header("Docker-Upload-UUID", uuid);
				response.status(202);
				return "";
			});

			// Completed Upload
			Spark.put(path + "/v2/:name/blobs/uploads/:uuid", (request, response) -> {

				String repo = "docker"; // request.params(":repo");
				String name = request.params(":name");
				String uuid = request.params(":uuid");

				// ?digest=<digest>
				String digest = request.queryParams("digest");

				DockerBlob blob = DockerBlobFactory.get(repo, name, uuid);
				blob.complete(digest);
				long length = blob.getLength();

				// upload created
				String location = "/v2/" + name + "/blobs/" + digest;
				response.header("Location", location);
				response.header("Content-Length", String.valueOf(length));
				response.header("Docker-Content-Digest", digest);
				response.status(201);
				return "";
			});

			// Canceling an Upload
			Spark.delete(path + "/v2/:name/blobs/uploads/:uuid", (request, response) -> {

				String repo = "docker"; // request.params(":repo");
				String name = request.params(":name");
				String uuid = request.params(":uuid");

				DockerBlob blob = DockerBlobFactory.get(repo, name, uuid);
				blob.cancel();

				return "";
			});

		}
	}

	private static int getManifestVersion(String contentType) {
		int version = 1;
		if (contentType != null && contentType.contains("manifest.v2")) {
			version = 2;
		}
		return version;
	}

	public static ModelAndView getDeploy(Request request, Response response) {
		Map<String, Object> model;
		model = NavBar.attributes("deploy", request);

		// get type from request parameters
		ArtifactType type = getType();
		model.put("type", type);

		try {
			IOptionsServer p4 = ConnectionSession.get(request, response);
			model.put("depots", DepotFactory.listDeployable(p4, type));
			p4.disconnect();
		} catch (P4JavaException e) {
			return ErrorModel.error(e);
		}

		return new ModelAndView(model, "/deploy/page-upload.html");
	}

	public static String postDeploy(Request request, Response response) throws Exception {
		// get type from request parameters
		ArtifactType type = getType();

		ArtifactInterface artifact = PluginArtifactFactory.getArtifact(type, request, response);
		List<String> upload = artifact.populateMultiPartPost();
		int pending = artifact.updatePending(upload);
		artifact.disconnect();

		// pending change
		return "/submit/" + type.getName() + "?pending=" + pending;
	}

}
